#!usr/bin/env
###
#Extra functions for random things
###
#TODO: everything needs a 
#returns
#   `type`

import numpy as np



#==================================================
#degrees
#==================================================
def degrees(theta: float) -> float:
    """Convert from radians to degrees
    
    Parameters
    ----------
    theta : `float`
        The angle in radians \n
    
    Examples
    --------
    >>> import gas_dynamics as gd
    >>> rad = gd.degrees(theta=3.14159)
    >>> rad
    179.99984796050427
    >>> 
    """

    return theta * 180/np.pi



#==================================================
#radians
#==================================================
def radians(theta: float) -> float:
    """Convert from degrees to radians
    
    Parameters
    ----------
    theta : `float`
        The angle in degrees\n
    
    Examples
    --------
    >>> import gas_dynamics as gd
    >>> deg = gd.radians(theta=180)
    >>> deg
    3.141592653589793
    >>> 
    """

    return theta * np.pi/180



#==================================================
#sine degrees
#==================================================
def sind(theta: float) -> float:
    """Sine given degrees
    
    Parameters
    ----------
    theta : `float`
        The angle in degrees
    
    Examples
    --------
    >>> import gas_dynamics as gd
    >>> sine = gd.sind(theta=45)
    >>> sine
    0.7071067811865476
    >>>
    """
    
    theta_radians = theta * np.pi / 180
    return np.sin(theta_radians)
    


#==================================================
#arc sine degrees
#==================================================
def arcsind(sin: float) -> float:
    """Return the arcsine in degrees
    
    Parameters
    ----------
    sin : `float`
        The sine of theta
    
    Examples
    --------
    >>> import gas_dynamics as gd
    >>> theta = gd.arcsind(sin = .7071)
    >>> theta
    44.99945053347443
    >>>
    """

    theta_rad = np.arcsin(sin)
    theta = degrees(theta_rad)
    return theta



#==================================================
#cosine degrees
#==================================================
def cosd(theta: float) -> float:
    """Cosine given degrees
    
    Parameters
    ----------
    theta : `float`
        The angle in degrees
    
    Examples
    --------
    >>> import gas_dynamics as gd
    >>> cosine = gd.cosd(theta=45)
    >>> cosine
    0.7071067811865476
    >>>
    """

    theta_radians = theta * np.pi / 180
    return np.cos(theta_radians)



#==================================================
#arc cosine degrees
#==================================================
def arccosd(cos: float) -> float:
    """Return the arccosine in degrees
    
    Parameters
    ----------
    cos : `float`
        The cosine of theta
    
    Examples
    --------
    >>> import gas_dynamics as gd
    >>> theta = gd.arccosd(.7071) 
    >>> theta
    45.00054946652557
    >>>
    """

    theta_rad = np.arccos(cos)
    theta = degrees(theta_rad)
    return theta



#==================================================
#tangent degrees
#==================================================
def tand(theta: float) -> float:
    """Tangent given degrees
    
    Parameters
    ----------
    theta : `float`
        The angle in degrees
    
    Examples
    --------
    >>> tan = gd.tand(theta=45)
    >>> tan
    0.9999999999999999
    >>>  
    """

    theta_rad = theta * np.pi / 180
    return np.tan(theta_rad)



#==================================================
#arc tangent degrees
#==================================================
def arctand(tan: float) -> float:
    """Return the arctangent in degrees
    
    Parameters
    ----------
    tan : `float`
        The tangent of theta
    
    Examples
    --------
    >>> theta = gd.arctand(1)
    >>> theta
    45.0
    >>>
    """
    
    theta_rad = np.arctan(tan)
    theta = degrees(theta_rad)
    return theta



#==================================================
#are to diameter
#TODO: deprecate and split into two
#==================================================
def area_dia(dia=[], area=[]):
    """
    Notes
    -----
    
    Given area or diameter, return the unknown
    
    Parameters
    ----------
    dia: Diameter \n
    area: Area \n
    
    Examples
    --------
    >>> import gas_dynamics as gd
    >>> area = gd.area_dia(dia=1)    
    >>> area
    0.7853981633974483
    >>> dia = gd.area_dia(area=area)
    >>> dia 
    1.0
    >>>
    """

    if area == []:
        area = np.pi *dia**2 / 4
        return area
    if dia == []:
        dia = (area * 4 / np.pi )**.5
        return dia



#==================================================
#linear interpolate
#==================================================
def lin_interpolate(x, x0, x1, y0, y1):
    """Linear interpolation formula
    
    Notes
    -----
    Given two x values and their corresponding y values, interpolate for
    an unknown y value at x
    
    Parameters
    ----------
    x : `float`
        The x value at which the y value is unknown \n
    x0 : `float`
        The x value of the first known pair \n
    x1 : `float`
        The x value of the second known pair \n
    y0 : `float`
        The y value of the first known pair \n
    y1 : `float`
        The y value of the second known pair \n
    
    Examples
    --------
    >>> import gas_dynamics as gd
    >>> y = gd.lin_interpolate(x=5, x0=4, x1=6, y0=25, y1=33)
    >>> y
    29.0
    >>>
    """

    y = y0 + (x-x0) * (y1-y0)/(x1-x0)
    return y


#!usr/bin/env
#Equations, plots, and tables for working with shocks.
#Included are functions for solving problems related to properties across
#normal and oblique shocks, from stagnation relations and changes to
#conditions to flow deflections. Tables can be made for any gas and its respective 
#ratio of specific heats, as well as plots and charts of relationships.
#
#  Typical usage example:
#  Generate a shock tabgle for argon (two columns omitted for readability)
#  >>> gd.shock_tables(range=[1,2], step=.1, gas=argon) 
#  Normal Shock Parameters for γ = 1.67
#  M: 1.00   |   M2: 1.0000   |    p2/p1: 1.0000   |    T2/T1: 1.0000   |   
#  M: 1.10   |   M2: 0.9131   |    p2/p1: 1.2627   |    T2/T1: 1.0985   |  
#  M: 1.20   |   M2: 0.8463   |    p2/p1: 1.5504   |    T2/T1: 1.1956   |  
#  M: 1.30   |   M2: 0.7935   |    p2/p1: 1.8631   |    T2/T1: 1.2933   |   
#  M: 1.40   |   M2: 0.7509   |    p2/p1: 2.2009   |    T2/T1: 1.3934   |  
#  M: 1.50   |   M2: 0.7158   |    p2/p1: 2.5637   |    T2/T1: 1.4968   |   
#  M: 1.60   |   M2: 0.6866   |    p2/p1: 2.9515   |    T2/T1: 1.6042   |  
#  M: 1.70   |   M2: 0.6620   |    p2/p1: 3.3643   |    T2/T1: 1.7162   |  
#  M: 1.80   |   M2: 0.6410   |    p2/p1: 3.8021   |    T2/T1: 1.8331   |   
#  M: 1.90   |   M2: 0.6229   |    p2/p1: 4.2649   |    T2/T1: 1.9552   |   
#  M: 2.00   |   M2: 0.6073   |    p2/p1: 4.7528   |    T2/T1: 2.0827   |  
#
#Copyright 2020 by Fernando A de la Fuente
#All rights reserved


import numpy as np
from scipy.optimize import fsolve
import matplotlib.pyplot as plt
from matplotlib.ticker import (MultipleLocator, FormatStrFormatter,AutoMinorLocator)
from gas_dynamics.extra import ( radians, degrees, sind, arcsind, cosd, arccosd, tand, arctand, lin_interpolate )
from gas_dynamics.fluids import fluid, air



#==================================================
#shock_mach
#fix runtime double scalars warning
#==================================================
def shock_mach(M: float, gas=air) -> float:
    """Returns the Mach number after a standing normal shock

    Notes
    -----
    Given a starting Mach number M1 and the ratio of specific heats,
    return the Mach number M2 that immediately follows the shock.
    The default fluid is air.

    Parameters
    ----------
    M : `float`
        The Mach number before the shock\n
    gas : `fluid`
        A user defined fluid object. Default is air \n

    Returns
    -------
    float
        The Mach number after the shock\n

    Examples
    --------
    >>> M2 = gd.shock_mach(M1=1.5) 
    >>> M2
    0.7010887416930995
    >>> 
    """
    
    gamma = gas.gamma

    #TODO: still getting runtime double scalars, filter out for this better
    M2 = ((M**2 + 2/(gamma-1)) / ((2*gamma / (gamma-1)) * M**2 - 1))**.5
    return M2



#==================================================
#shock_mach_before
#==================================================
def shock_mach_before(M: float, gas=air) -> float:
    """Returns the Mach number before a standing normal shock
    
    Notes
    -----
    Given the Mach number after the shock and the ratio of specific heats,
    return the Mach number that immediately precedes the shock.
    Default fluid is air.
    
    Parameters
    ----------
    M : `float`
        The Mach number after the shock\n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    
    Returns
    -------
    float
        The mach number before the shock\n
    
    Examples
    --------
    >>> import gas_dynamics as gd 
    >>> M2 = 0.7010887416930995
    >>> M1 = gd.shock_mach(M2)
    >>> M1
    1.4999999999999998
    >>>
    """

    gamma = gas.gamma
    M1 = ((-2/(gamma-1) -M**2 ) / (1- ((2*gamma)/(gamma-1))*M**2))**.5
    return M1



#==================================================
#prandtl_meyer_mach
#added fluid class, split into two funcs
#==================================================
def shock_pressure_ratio(M: float, gas=air) -> float:
    """Returns the pressure ratio after a standing normal shock for a given Mach number
    
    Notes
    -----
    Given a starting Mach number and a ratio of specific heats, this
    function returns the ratio of pressure 2 over pressure 1 across a
    standing normal shock. 

    Parameters
    ----------
    M : `float`
        The starting Mach number \n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    
    Returns
    -------
    float
        The pressure ratio p2/p1\n

    Examples
    --------
    >>> import gas_dynamics as gd
    >>> p2_p1 = gd.shock_pressure_ratio(M=1.5) 
    >>> p2_p1   
    2.4583333333333335
    >>>
    """

    gamma = gas.gamma
    p2_p1 = 2*gamma / (gamma+1) * M**2 - (gamma-1)/(gamma+1)
    return p2_p1



#==================================================
#shock mach from pressure ratio
#==================================================
def shock_mach_from_pressure_ratio(p2_p1: float, gas=air) -> float:
    """Returns the mach number across a standing normal shock given a pressure ratio
    
    Notes
    -----
    Given the ratio of pressure behind the shock over pressure before
    the shock and the ratio of specific heats, this function returns the 
    Mach number before the shock. Default fluid is air.
    
    Parameters
    ----------
    p2_p1 : `float`
        The pressure ratio\n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    
    Returns
    -------
    float
        The mach number prior the shock\n
    
    Examples
    --------
    >>> import gas_dynamics as gd 
    >>> p2_p1 = 6
    >>> M1 = gd.shock_mach_from_pressure_ratio(p2_p1)
    >>> M1
    2.29906813420444
    >>> 
    """

    gamma = gas.gamma
    M = ((gamma+1)/(2*gamma) * (p2_p1 + (gamma-1)/(gamma+1)) )**.5
    return M



#==================================================
#shock_temperature_ratio
#==================================================
def shock_temperature_ratio(M: float, gas=air) -> float:
    """Returns the temperature ratio after a standing normal shock for a given Mach number
    
    Notes
    -----
    Given a starting Mach number and a ratio of specific heats, this function
    returns the ratio of temperature 2 over temperature 1 across a standing normal
    shock. Default fluid is air.
    
    Parameters
    ----------
    M : `float`
        The starting Mach number \n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    
    Returns
    -------
    float
        The temperature ratio T2/T1\n

    Examples
    --------
    >>> import gas_dynamics as gd
    >>> T2_T1 = gd.shock_temperature_ratio(M=1.5)
    >>> T2_T1
    1.320216049382716
    >>>
    """

    gamma = gas.gamma
    term1 = (1 + (gamma-1)/2 * M**2)
    term2 = (2*gamma)/(gamma-1) * M**2 -1
    term3 = (gamma+1)**2 / (2*(gamma-1)) * M**2
    t2_t1 = (term1 * term2) / term3
    return t2_t1



#==================================================
# shock_dv_a
#need to add examples, maybe change name?
#TODO: ADD EXAMPLES
#==================================================
def shock_dv_a(M: float, gas=air) -> float:
    """Returns change in velocity over the local speed of sound after a normal shock.
    
    Notes
    -----
    Given a starting Mach number and a ratio of specific heats, this function
    returns the velocity change across a standing normal shock divided by the
    local speed of sound. Default fluid is air
    
    Parameters
    ----------
    M : `float`
        The starting Mach number \n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    
    Returns
    -------
    float
        The change in Mach number\n

    Examples
    --------

    """

    gamma = gas.gamma
    dv_a = 2/(gamma+1) * (M**2 -1)/ M
    return dv_a



#==================================================
#shock_stagnation ratio
#==================================================
def shock_stagnation_ratio(M: float, gas=air) -> float:
    """Returns stagnation pressure ratio after a normal shock.
    
    Notes
    -----
    Given a starting Mach number and a ratio of specific heats, this function
    returns the ratio of stagnation presure 2 over stagnation pressure 1
    across a standing normal shock. Default fluid is air.
    
    Parameters
    ----------
    M : `float`
        The starting Mach number \n
    gas : `fluid`
        A user defined fluid object. Default is air \n    

    Returns
    -------
    float
        The stagnation pressure ratio pt2/pt1\n

    Examples
    --------
    >>> pt2_pt1 = gd.shock_stagnation_ratio(M=2)
    >>> pt2_pt1
    0.7208738614847454
    >>>
    """

    gamma = gas.gamma
    term1 = (gamma+1)/2*M**2
    term2 = 1 + (gamma-1)/2 * M**2
    term3 = (term1/term2) ** (gamma/(gamma-1))
    term4 = (2*gamma / (gamma+1) * M**2 - ((gamma-1)/(gamma+1)))**(1/(1-gamma))
    return term3 * term4



#==================================================
#shock_tables
# examples!

#==================================================
def shock_tables(range=[1,5], step=.01, gas=air) -> str:
    """Returns shock tables for a range of Mach numbers.
    
    Notes
    -----
    Given a range of Mach numbers and a ratio of specific heats, generate
    the standing normal shock tables for every incremental Mach number
    in between.
    
    Parameters
    ----------
    range : `list`
        The starting and ending Mach # in a list, ie: [1,5]. \n
    step : `float`
        The step size for the tables. \n
    gas : `fluid`
        A user defined fluid object. Default is air \n    
    
    Returns
    -------
    str
        The shock table\n

    Examples
    --------
    >>> import gas_dynamics as gd
    >>> gd.shock_tables(range=[1,2], step=.1)
    Normal Shock Parameters for γ = 1.4
    M: 1.00   |   M2: 1.0000   |    p2/p1: 1.0000   |    T2/T1: 1.0000   |   dV/a: 0.0000   |   pt2/pt1: 1.000000
    M: 1.10   |   M2: 0.9118   |    p2/p1: 1.2450   |    T2/T1: 1.0649   |   dV/a: 0.1591   |   pt2/pt1: 0.998928
    M: 1.20   |   M2: 0.8422   |    p2/p1: 1.5133   |    T2/T1: 1.1280   |   dV/a: 0.3056   |   pt2/pt1: 0.992798
    M: 1.30   |   M2: 0.7860   |    p2/p1: 1.8050   |    T2/T1: 1.1909   |   dV/a: 0.4423   |   pt2/pt1: 0.979374
    M: 1.40   |   M2: 0.7397   |    p2/p1: 2.1200   |    T2/T1: 1.2547   |   dV/a: 0.5714   |   pt2/pt1: 0.958194
    M: 1.50   |   M2: 0.7011   |    p2/p1: 2.4583   |    T2/T1: 1.3202   |   dV/a: 0.6944   |   pt2/pt1: 0.929787
    M: 1.60   |   M2: 0.6684   |    p2/p1: 2.8200   |    T2/T1: 1.3880   |   dV/a: 0.8125   |   pt2/pt1: 0.895200
    M: 1.70   |   M2: 0.6405   |    p2/p1: 3.2050   |    T2/T1: 1.4583   |   dV/a: 0.9265   |   pt2/pt1: 0.855721   
    M: 1.80   |   M2: 0.6165   |    p2/p1: 3.6133   |    T2/T1: 1.5316   |   dV/a: 1.0370   |   pt2/pt1: 0.812684
    M: 1.90   |   M2: 0.5956   |    p2/p1: 4.0450   |    T2/T1: 1.6079   |   dV/a: 1.1447   |   pt2/pt1: 0.767357
    M: 2.00   |   M2: 0.5774   |    p2/p1: 4.5000   |    T2/T1: 1.6875   |   dV/a: 1.2500   |   pt2/pt1: 0.720874
    >>> 
    """

    Mach_min = range[0]
    if Mach_min < 1:
        Mach_min = 1

    Mach_max = range[1]

    mach_nums = [i for i in np.arange(Mach_min,Mach_max+step,step)]
    M2 = [shock_mach(M=i, gas=gas)for i in mach_nums]
    p2_p1 = [shock_pressure_ratio(M=i, gas=gas)for i in mach_nums]
    T2_T1 = [shock_temperature_ratio(M=i, gas=gas) for i in mach_nums]
    dv_a = [shock_dv_a(M=i, gas=gas) for i in mach_nums]
    pt2_pt1 = [shock_stagnation_ratio(M=i, gas=gas) for i in mach_nums]
    gamma = gas.gamma

    labl = '\u03B3 = ' + str(gamma)
    print("Normal Shock Parameters for " + gas.name + ", " + labl)
    for index, num in enumerate(mach_nums):
        print("M: " + f"{num:.2f}" + "   |"+"   M2: " + f"{M2[index]:.4f}" + "   | " + "   p2/p1: " + f"{p2_p1[index]:.4f}" + "   | "+"   T2/T1: " + f"{T2_T1[index]:.4f}" + "   |"+"   dV/a: " + f"{dv_a[index]:.4f}"+ "   |"+"   pt2/pt1: " + f"{pt2_pt1[index]:.6f}" )
    print("\n \n \n")



#==================================================
#shock_flow_deflection
# good! added fluid class
#==================================================
def shock_flow_deflection(M: float, theta: float, gas=air) -> float:
    """Returns flow deflection angle from Mach number and oblique shock angle
    
    Notes
    -----
    Given the Mach number prior to the oblique shock, the angle of the oblique
    shock in degrees, and the ratio of specific heats, this function returns
    the angle that the flow is turned. Default fluid is air.
    
    Parameters
    ----------
    M : `float`
        The Mach number before the shock \n
    theta : `float`
        The shock angle in degrees \n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    
    Returns
    -------
    float
        The deflection angle\n

    Examples
    --------
    >>> import gas_dynamics as gd
    >>> deflect = gd.shock_flow_deflection(M=2, theta = 22.5)
    >>> deflect
    -10.856560004139958
    >>>        
    """

    gamma = gas.gamma
    dirac = arctand( 2 * 1/tand(theta) * (M**2 * sind(theta)**2 - 1 ) / (M**2 * (gamma + cosd(2*theta)) + 2 ))    
    return dirac



#==================================================
#shock_angle
# good! added fluid class
#==================================================
def shock_angle(M: float, dirac: float, gas=air) -> float:
    """Return the shock angle given the Mach number prior to the shock and the deflection angle
    
    Notes
    -----
    Given the Mach number prior to the oblique shock, the angle of the flow
    deflection, and the ratio of specific heats, this functions returns the
    angle that is formed by the shock. Default ratio of specific heats is
    for air
    
    Parameters
    ----------
    M : `float`
        The Mach number before the shock \n
    dirac : `float`
        The flow deflection angle in degrees\n
    gas : `fluid`
        A user defined fluid object. Default is air \n    
    
    Returns
    -------
    float
        The shock angle\n

    Examples
    --------
    >>> import gas_dynamics as gd
    >>> shocks = gd.shock_angle(M=2, dirac = -10) 
    >>> shocks
    [23.014012220565785, 96.29991962425305]
    >>> 
    """

    gamma = gas.gamma
    def func(theta, M=M, dirac=dirac, gamma=gamma):
        zero = 2 * 1/tand(theta) * (M**2 * sind(theta)**2 - 1 ) / (M**2 * (gamma + cosd(2*theta)) + 2 ) - tand(dirac)
        return zero

    weak = fsolve(func, x0=0.001, args=(M, dirac, gamma))
    strong = fsolve(func, x0=90, args=(M, dirac, gamma))
    shock_angles = [weak[0], strong[0]]
    return shock_angles



#==================================================
#shock_mach_given angles
#==================================================
def shock_mach_given_angles(theta: float, dirac: float, gas=air) -> float:
    """Return the Mach number given the shock angle and flow deflection
    
    Notes
    -----
    Given the angle of the shock and the angle that the flow has turned,
    return the mach number that preceded the shock.

    Parameters
    ----------
    theta : `float`
        The shock angle in degrees \n
    dirac : `float`
        The flow deflection angle in degrees \n
    gas : `fluid`
        A user defined fluid object. Default is air \n    
    
    Returns
    -------
    float
        The mach number\n

    Examples
    --------
    >>> import gas_dynamics as gd
    >>> M = gd.shock_mach_given_angles(theta=22.5, dirac=10) 
    >>> M
    3.9293486839798955
    >>>
    """

    gamma = gas.gamma
    def func(M, theta=theta, dirac=dirac, gamma=gamma):
        '''
        Zero function for solving for the mach number
        '''
        zero = 2 * 1/tand(theta) * (M**2 * sind(theta)**2 - 1 ) / (M**2 * (gamma + cosd(2*theta)) + 2 ) - tand(dirac)
        return zero

    sol = fsolve(func, x0=0.001, args=(theta, dirac, gamma))
    return sol[0]



#==================================================
#prandtl_meyer_turn
#need to add examples
#==================================================
def prandtl_meyer_turn(M: float, gas=air) -> float:
    """Returns the angle through which a flow has turned to reach a Mach number
    
    Notes
    -----
    Given a Mach number and ratio of specific heats, calculate the angle of a turn
    through which a flow has traversed to reach the Mach number given, from a  Mach number
    of 1. Also known as the Prandtl-Meyer function. Default fluid is air.
    
    Parameters
    ----------
    M : `float`
        The Mach number \n
    gas : `fluid`
        A user defined fluid object. Default is air \n    
    
    Returns
    -------
    float
        The angle through which the flow has turned\n

    Examples
    --------

    """

    gamma = gas.R
    nu = ((gamma+1)/(gamma-1))**.5 * arctand(((M**2-1)*(gamma-1)/(gamma+1))**.5) - arctand((M**2-1)**.5)
    return nu



#==================================================
#prandtl_meyer_mach
#need to add examples
#==================================================
def prandtl_meyer_mach(nu: float, gas=air) -> float:
    """Returns the Mach number given an angle through which the flow has turned
    
    Notes
    -----
    Given a smooth turn through which a flow has turned and the ratio of specific
    heats, return the Mach number after the turn.

    Parameters
    ----------
    nu : `float`
        The turn angle \n    
    gas : `fluid`
        A user defined fluid object. Default is air \n    

    Returns
    -------
    float
        The mach number\n

    Examples
    --------
    >>> 
    >>> 
    """
    
    def get_mach(M: float, nu=nu, gas=gas) -> float:
        return prandtl_meyer_turn(M, gas=gas) - nu
    
    sol = fsolve(get_mach, x0=1.5, args=(nu, gas))
    return sol[0]



#==================================================
#shock_oblique_charts
#need to add examples, parameters, descriptions
#==================================================
def shock_oblique_charts(Mach_max=6, gas=air, points=40000, dark=True):
    """Generate 2-D Oblique Shock Charts
    
    Notes
    -----
    Displays two plots,
    1) Mach number versus oblique shock wave angle and the corresponding
    deflection angles for each case. 2) Mach number versus resulting Mach
    number after an oblique shock and the corresponding deflection angles
    for each case. Default ratio of specific heats is for air.
    
    Parameters
    ----------
    Mach_max : `float`
        The upper limit Mach number for the chart \n
    gas : `fluid`
        A user defined fluid object. Default is air \n    
    points : `int`
        The number of points to evaluate on the mesh\n
    dark : `bool`
        Dark mode for the plots. Default is true.\n
    
    Examples
    --------
    >>> gd.shock_oblique_charts()
    >>>
    """

    n = 1000
    gamma = gas.gamma

    if dark == True:
        plt.style.use('dark_background')
        grid_color = 'w'
    else: 
        plt.style.use('default')
        grid_color = 'k'


    theta = np.linspace(.001,90, n)
    mach = np.linspace(1,Mach_max,n)
    MACH,THETA = np.meshgrid(mach,theta)
    dirac = shock_flow_deflection(M=MACH, theta=THETA, gas=gas)
    fig, (ax1,ax2) = plt.subplots(1,2)
    levels=[0, 5,10,15,20,25,30,35,40]
    h = ax1.contour(mach,theta,dirac,levels=levels,cmap='tab10')
    ax1.clabel(h, inline =1, fontsize=10)
    minor_ticks_mach = np.arange(1,Mach_max+.1,.1)
    minor_ticks_theta = np.arange(0,91,1)
    ax1.set_yticks(minor_ticks_theta, minor=True)
    ax1.set_xticks(minor_ticks_mach, minor=True)
    ax1.grid(which='major', color=grid_color, linestyle = '--', alpha=.5,)
    ax1.grid(which='minor', color=grid_color, linestyle = '--', alpha=.1)
    ax1.set(xlabel = 'Mach #')
    ax1.set(ylabel = 'Oblique Shock Wave Angle')

    n = round(points**.5)
    n = 750

    total, counter = n**2, 0 
    mach_before = np.linspace(1,Mach_max, n)
    mach_after = np.linspace(.001,Mach_max, n)
    dirac = np.zeros((n,n))
    for row, m1 in enumerate(mach_before):
        percent =counter / total * 100
        print(' %0.1f%% complete' %percent)
        for col, m2 in enumerate(mach_after):
            dirac[col][row] = dirac_from_machs(M1=m1, M2=m2, gas=gas)
            counter += 1

    h = ax2.contour(mach_before, mach_after, dirac , levels=levels, cmap='tab10')
    ax2.clabel(h, inline = 1, fontsize=10)
    minor_ticks_mach_before = np.arange(0,Mach_max+.1,.1)
    minor_ticks_mach_after = np.arange(0,Mach_max,.1)
    ax2.set_yticks(minor_ticks_mach_before, minor=True)
    ax2.set_xticks(minor_ticks_mach_after, minor=True)
    ax2.grid(which='major',color=grid_color,linestyle = '--', alpha=.5,)
    ax2.grid(which='minor',color=grid_color,linestyle = '--', alpha=.1)
    ax2.set(xlabel = 'Mach # Before')
    ax2.set(ylabel = 'Mach # After')

    #plot title stuff
    string = 'Oblique Shock and Flow Deflection Angles for ' + gas.name
    fig.suptitle(string)
    fig.tight_layout(pad=2.0)
    plt.show()



#==================================================
#dirac from machs
##TODO: better numerical method for this
##TODO: examples
#==================================================
def dirac_from_machs(M1: float, M2: float, gas=air) -> float:
    """Return the flow deflection angle given the mach number before
    and after the oblique shock
    
    Notes
    -----
    Given two mach numbers, iteratively solve for the shock angle and
    flow deflection to satisfy the system.

    Parameters
    ----------
    M1 : `float`
        The initial mach number \n
    M2 : `float`
        The mach number after the event \n
    gas : `fluid`
        A user defined fluid object. Default is air \n 

    Returns
    -------
    float
        The flow deflection angle\n

    Examples
    --------
    >>> import gas_dynamics as gd
    >>> flow_deflect = gd.dirac_from_machs(M1=3, M2=1.5)
    >>> flow_deflect
    28.589471710648365 
    >>>
    """   
    
    if M2 >= M1:
        return 0
    if shock_mach(M1) > M2:
        return 0

    def zero(theta: float, M1:float, M2:float, gas=gas) -> float:
        """Local function for testing different shock angles to see if they work.
        
        """

        M1n = M1 * sind(theta)
        M2n = shock_mach(M1n)
        dirac = shock_flow_deflection(M=M1, theta=theta, gas=gas)
        M2_prime = M2n / sind(theta-dirac)
        zero = M2_prime - M2
        return zero

    thetas = np.linspace(0, 90, 100)
    for num, theta in enumerate(thetas[:-1]):
        zero1 = zero(thetas[num], M1=M1, M2=M2, gas=gas)
        #zerolist.append(zero1)
        if zero1 < 0:
            zero2 = zero(thetas[num-1], M1=M1, M2=M2, gas=gas)
            theta = lin_interpolate(0, zero1, zero2, thetas[num], thetas[num-1])
            return shock_flow_deflection(M=M1, theta=theta, gas=gas)


#!usr/bin/env
###
#The fluid class and some common fluids and their properties in metric and standard
###

#==================================================
#fluid
#==================================================
class fluid:
    """A class to represent the fluid and its properties

    Attributes
    ----------
    gamma : `float`
        The ratio of specific heats \n
    R : `float`
        The gas constant for the fluid \n
    units : `str`
        The unit system defining the gas constant \n

    Methods
    -------
    No methods at this time

    Examples
    --------
    >>> import gas_dynamics as gd
    >>> methane = gd.fluid('methane', 1.3, 518.2)
    >>> methane.gamma
    1.3
    >>> methane.R
    518.2
    >>> methane.units
    'metric'
    Conversely we can set the units
    >>> methane = fluid('methane', 1.3, 0.1238, units = 'btu / lbm-R')
    >>> methane.gamma
    1.3
    >>> methane.R
    0.1238
    >>> methane.units
    'btu / lbm-R'

    """

    def __init__(self, name: str, gamma: float, R: float, units='metric'):
        """Construct the necessary attributes for the fluid object

        Parameters
        ----------
        name : `str`
            The name of the fluid\n
        gamma : `float`
            The ratio of specific heats \n
        R : `flaot`
            The gas constant for the fluid \n
        units : `str`
            The units being used for the gas constant. Default is metric \n

        """
        self.name = name
        self.gamma = gamma
        self.R = R
        self.units = units



#Initialize some fluids in the metric system
#==================================================
air = fluid(name='Air', gamma=1.4, R=286.9, units='J / kg-K')

argon = fluid(name='Argon', gamma=1.67, R=208, units='J / kg-K')

CO2 = fluid(name='Carbon Dioxide', gamma=1.29, R=189, units='J / kg-K')

CO = fluid(name='Carbon Monoxide', gamma=1.4, R=297, units='J / kg-K')

hydrogen = fluid(name='Hydrogen', gamma=1.41, R=4120, units='J / kg-K')

helium = fluid(name='Helium', gamma=1.67, R=2080, units='J / kg-K')

methane = fluid(name='Methane', gamma=1.32, R=519, units='J / kg-K')

nitrogen = fluid(name='Nitrogen', gamma=1.4, R=296, units='J / kg-K')

O2 = fluid(name='Oxygen', gamma=1.4, R=260, units='J / kg-K')

water = fluid(name='water', gamma=1.33, R=461, units='J / kg-K')


#and in the british standard (Btu / lbm-R)
#============================================================
air_us = fluid(name='Air', gamma=1.4, R=53.3, units='Btu / lbm-R')

argon_us = fluid(name='Argon', gamma=1.67, R=38.7, units='Btu / lbm-R')

CO2_us = fluid(name='Carbon Dioxide', gamma=1.29, R=35.1, units='Btu / lbm-R')

CO_us = fluid(name='Carbon Monoxide', gamma=1.4, R=55.2, units='Btu / lbm-R')

hydrogen_us = fluid(name='Hydrogen', gamma=1.41, R=766, units='Btu / lbm-R')

helium_us = fluid(name='Helium', gamma=1.67, R=386, units='Btu / lbm-R')

methane_us = fluid(name='Methane', gamma=1.32, R=96.4, units='Btu / lbm-R')

nitrogen_us = fluid(name='Nitrogen', gamma=1.4, R=55.1, units='Btu / lbm-R')

O2_us = fluid(name='Oxygen', gamma=1.4, R=48.3, units='Btu / lbm-R')

water_us = fluid(name='water', gamma=1.33, R=85.7, units='Btu / lbm-R')


from unittest import TestCase
import gas_dynamics as gd


#==================================================
#Build a whole test suite for all funcs, default arguments, and other gasses
#==================================================


#==================================================
#sonic_velocity test
#==================================================
class test_sonic_velocity(TestCase):
    #test we get a number
    def test_is_float(self):
        a = gd.sonic_velocity()
        self.assertTrue(isinstance(a, float))


#==================================================
#stagnation pressure test
#==================================================
class test_stagnation_pressure(TestCase):
    #test we get a number
    def test1(self):
        p = gd.stagnation_pressure(pt=2, M=1)
        self.assertTrue(isinstance(p, float))
        
        pt = gd.stagnation_pressure(p=p, M=1)
        self.assertTrue(isinstance(pt, float))

        M = gd.stagnation_pressure(pt=pt, p=p)
        self.assertTrue(isinstance(M, float))

        self.assertNotEqual(p,pt)
        self.assertNotEqual(p,M)
        self.assertNotEqual(pt,M)        
        

#==================================================
#stagnation temperature test
#==================================================
class test_stagnation_temperature(TestCase):
    #test we get a number
    def test1(self):
        T = gd.stagnation_temperature(Tt=2, M=1)
        self.assertTrue(isinstance(p, float))
        
        Tt = gd.stagnation_temperature(T=T, M=1)
        self.assertTrue(isinstance(Tt, float))

        M = gd.stagnation_pressure(Tt=Tt, T=T)
        self.assertTrue(isinstance(M, float))

        self.assertNotEqual(T,Tt)
        self.assertNotEqual(T,M)
        self.assertNotEqual(Tt,M)        



#!usr/bin/env

#import isentropic functions and stagnation relations
from gas_dynamics.standard.standard import ( 
  sonic_velocity,
  stagnation_pressure,
  stagnation_temperature,
  stagnation_pressure_ratio,
  stagnation_temperature_ratio,
  stagnation_density_ratio,
  stagnation_ratio,
  stagnation_ratio_table,
  mach_from_pressure_ratio,
  mach_from_temperature_ratio,
  pressure_from_mach_ratio,
  temperature_from_mach_ratio,
  entropy_produced,
  mach_area_ratio_choked,
  mach_area_ratio,
  mach_from_area_ratio,
  mass_flux_max,
  mass_flux,
  plot_stagnation_ratios )

#import shock functions
from gas_dynamics.shocks.shocks import (
  shock_mach,
  shock_mach_before,
  shock_pressure_ratio,
  shock_mach_from_pressure_ratio,
  shock_temperature_ratio,
  shock_dv_a, 
  shock_stagnation_ratio, 
  shock_flow_deflection, 
  shock_angle, 
  shock_mach_given_angles, 
  prandtl_meyer_turn, 
  prandtl_meyer_mach, 
  shock_oblique_charts,
  shock_tables, 
  dirac_from_machs)

from gas_dynamics.fluids import fluid

#from .fanno import func1,
#from .rayleigh func 1


#!usr/bin/env
#Equations, plots, and tables for solving compresible flow problems.
#Functions range from stagnation relations to determining the mach
#number from changes in local properties to determining the maximum mass
#flux. Tables can be made for any gas and its respective ratio of specific
#heats, as well as plots and charts of relationships.

#  Typical usage example:
#  Generate isentropic flow stagnation relations for methane
#  >>> import gas_dynamics as gd
#  >>> gd.stagnation_ratio_table(range=[0,2], step=.2, gas=methane)
#  M: 0.000   |   P/Pt: 1.000    |    T/Tt: 1.000    |    A/A*: inf    |   rho/rho_t: 1.000
#  M: 0.200   |   P/Pt: 0.974    |    T/Tt: 0.994    |    A/A*: 2.988    |   rho/rho_t: 0.980
#  M: 0.400   |   P/Pt: 0.901    |    T/Tt: 0.975    |    A/A*: 1.600    |   rho/rho_t: 0.924
#  M: 0.600   |   P/Pt: 0.794    |    T/Tt: 0.946    |    A/A*: 1.192    |   rho/rho_t: 0.839
#  M: 0.800   |   P/Pt: 0.669    |    T/Tt: 0.907    |    A/A*: 1.039    |   rho/rho_t: 0.737
#  M: 1.000   |   P/Pt: 0.542    |    T/Tt: 0.862    |    A/A*: 1.000    |   rho/rho_t: 0.629
#  M: 1.200   |   P/Pt: 0.425    |    T/Tt: 0.813    |    A/A*: 1.032    |   rho/rho_t: 0.523
#  M: 1.400   |   P/Pt: 0.325    |    T/Tt: 0.761    |    A/A*: 1.121    |   rho/rho_t: 0.426
#  M: 1.600   |   P/Pt: 0.243    |    T/Tt: 0.709    |    A/A*: 1.267    |   rho/rho_t: 0.342
#  M: 1.800   |   P/Pt: 0.179    |    T/Tt: 0.659    |    A/A*: 1.474    |   rho/rho_t: 0.271
#  M: 2.000   |   P/Pt: 0.130    |    T/Tt: 0.610    |    A/A*: 1.754    |   rho/rho_t: 0.213
#
#Copyright 2020 by Fernando A de la Fuente
#All rights reserved

#TODO: everything needs a 
#returns
#   `type`


import numpy as np
from scipy.optimize import fsolve
import matplotlib.pyplot as plt
from gas_dynamics.fluids import fluid, air, methane, argon



#==================================================
#sonic_velocity
#fluid class and us standard option 11/6/2020
#==================================================    
def sonic_velocity(gas=air ,metric=True, T=273.15) -> float:
    """Returns the local speed of sound.
    
    Notes
    -----
    Given a ratio of specific heats, gas constant, and temperature
    this function returns the locoal speed of sound. Default fluid is air.
    
    Parameters
    ----------
    gas : `fluid`
        A user defined fluid object. Default is air \n
    metric : `bool`
        Use metric or US standard \n
    T : `float`
        The temperature \n

    Returns
    -------
    float
        The local speed of sound\n
    
    Examples
    --------
    >>> import gas_dynamics as gd
    >>> gd.sonic_velocity(air,T=500)
    448.2186966202994
    >>> gd.sonic_velocity(gas=Argon, T=300)

    """

    if metric == False:
        gc = 32.174
    else:
        gc = 1
    gamma, R = gas.gamma, gas.R
    a = ( gc * gamma * R *T )**.5
    return a



#==================================================
#stagnation_pressure
#implemented fluid class and string for output returned
#==================================================    
def stagnation_pressure(pt=None, M=None, p=None, gas=air, output=False) -> float:
    """Returns the stagnation pressure given pressure and Mach number.

    Notes
    -----
    Given a pressure, Mach number, and a ratio of specific heats return
    the stagnation pressure. Alternatively, provided two arguments
    the function will return the missing one. Default fluid is air.
    
    Parameters
    ----------
    pt : `float`
        The stagnation pressure.\n
    p : `float`
        The pressure.\n
    M : `float`
        The Mach number\n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    output : `bool`
        Print out a string to verify the output is the parameter desired \n
    
    Returns
    -------
    float
        The stagnation pressure, static pressure, or mach number\n

    Examples
    --------
    >>> import gas_dynamics as gd
    >>> pt = gd.stagnation_pressure(p=10, M=1)
    >>> pt
    18.92929158737854
    >>> M = gd.stagnation_pressure(p=10, pt=pt)
    >>> M
    1.0
    >>>
    """

    gamma = gas.gamma
    if pt == None:
        pt = p* ( 1 + (gamma-1)/2 * M**2)** (gamma/(gamma-1))
        if output == True:
            print('Returned stagnation pressure')
        return pt
   
    if M == None:
        M = (((pt/p)**((gamma-1)/gamma) -1 ) * 2/(gamma-1) ) ** .5
        if output == True:
            print('Returned Mach')
        return M

    if p == None:
        p = pt / ( 1 + (gamma-1)/2 * M**2)** (gamma/(gamma-1))
        if output == True:
            print('Returned pressure')
        return p



#==================================================
#stagnation_temperature
#implemented output string, fluid class, 
#==================================================    
def stagnation_temperature(T=None, Tt=None , M=None, gas=air, output=False) -> float :
    """Returns the stagnation temperature given temperature and Mach number.
    
    Notes
    -----
    Given a temperature, Mach number, and a ratio of specific heats 
    this function returns the stagnation temperature. Alternatively,
    provided two arguments the function will return the missing one.
    Default fluid is air.
    
    Parameters
    ----------
    Tt : `float`
        The stagnation temperature\n
    T : `float`
        The temperature\n
    M : `float`
        The Mach number\n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    output : `bool`
        Print out a string to verify the output is the parameter desired \n
    
    Returns
    -------
    float
        The stagnation temperature, temperature, or mach number\n

    Examples
    --------
    >>> Tt = gd.stagnation_temperature(T=300, M=1)
    >>> Tt
    360.0
    >>> M = gd.stagnation_temperature(T=300, Tt=Tt)
    >>> M 
    1.0
    >>>
    """

    gamma = gas.gamma
    if Tt == None:
        Tt = T * ( 1 + (gamma-1)/2 * M**2)
        if output == True:
            print('Returned stagnation temperature')
        return Tt

    if M == None:
        M = ((Tt /T - 1) * 2/(gamma-1))**.5
        if output == True:
            print('Returned Mach')
        return M
        
    if T == None:
        T = Tt/( 1 + (gamma-1)/2 * M**2)
        if output == True:
            print('Returned temperature')
        return T



#==================================================
#stagnation_pressure_ratio
#added fluid class, removed unnnecessary metric argument
#==================================================
def stagnation_pressure_ratio(M: float, gas=air) -> float:
    """Returns the pressure ratio of p / p_t
    
    Notes
    -----
    Given a Mach number and ratio of specific heats return the relation of
    pressure over stagnation pressure. Default fluid is air.
    
    Parameters
    ----------
    M : `float`
        The Mach number \n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    
    Returns
    -------
    float
        The stagnation pressure ratio\n

    Examples
    --------
    >>> import gas_dynamics as gd
    >>> p_pt = gd.stagnation_pressure_ratio(M=3)
    >>> p_pt
    0.027223683703862817
    >>>
    """

    gamma = gas.gamma
    denom = 1 + (gamma-1)/2 * M**2
    Pt_ratio = (1 / denom ) ** (gamma/(gamma-1))
    return Pt_ratio



#==================================================
#stagnation_temperature_ratio
#added fluid class
#==================================================
def stagnation_temperature_ratio(M: float, gas=air) -> float:
    """Returns the temperature ratio of T / T_t
    
    Notes
    -----
    Given a Mach number and ratio of specific heats return the relation of
    temperature over stagnation temperature.  Default fluid is air.
    
    Parameters
    ----------
    M : `float`
        The Mach number\n
    gas : `fluid`
        A user defined fluid object. Default is air\n
    
    Returns
    -------
    float
        The stagnation temperature ratio\n

    Examples
    --------
    >>> import gas_dynamics as gd
    >>> T_Tt = gd.stagnation_temperature_ratio(M=1.5)
    >>> T_Tt
    0.6896551724137931
    >>>
    """

    gamma = gas.gamma
    Tt_ratio = 1 / (1+(gamma-1)/2 *M**2)
    return Tt_ratio



#==================================================
#stagnation_density_ratio
#added fluid class
#==================================================
def stagnation_density_ratio(M: float, gas=air) -> float:
    """Returns the density ratio rho / rho_t
    
    Notes
    -----
    Given a Mach number and ratio of specific heats, return the relation
    of density over stagnation density. Default fluid is air.
    
    Parameters
    ----------
    M : `float`
        The Mach # \n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    
    Returns
    -------
    float
        The stagnation density ratio\n

    Examples
    --------
    >>> import gas_dynamics as gd
    >>> rho_rho_t = gd.stagnation_density_ratio(M=1.5) 
    >>> rho_rho_t
    0.39498444639115327
    >>>
    """

    gamma = gas.gamma
    rho_t_ratio = (1 / (1 + (gamma-1)/2 * M**2 )) ** (1 / (gamma-1))
    return rho_t_ratio



#==================================================
#stagnation ratio 
#TODO: make me
#added fluid object
#==================================================
def stagnation_ratio(M: float, gas=air) -> list:
    """Return stagnation pressure, temperature, density, and choked area ratio for a mach number
        
    Notes
    -----
    Given a mach number and the fluid, return the three stagnation ratios
    and the ratio of the area to the choked area. Default fluid is air

    Parameters
    ----------
    M : `float`
        The mach number
    gas : `fluid`
        A user defined fluid object. Default is air \n
    
    Returns
    -------
    list
        The stagnation pressure, stagnation temperature, stagnation
        density ratio, and choked area ratio.\n

    Examples
    --------
    >>> import gas_dynamics as gd

    """




#==================================================
#stagnation ratio tables
#added fluid object
#==================================================
def stagnation_ratio_table(range=[0,5], step=.1, gas=air) -> str:
    """Returns the isentropic flow tables in the given range.
    
    Notes
    -----
    Given a ratio of specific heats, print out the stagnation
    temperature ratio, stagnation pressure ratio, the area to 
    choked area ratio, and the stagnation density ratio for every
    incremental Mach number.
    
    Parameters
    ----------
    range : `list`
        The starting and ending Mach numbers in a list, ex: [0, 5] \n
    step : `float`
        The step size between min and max mach number \n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    
    Returns
    -------
    str
        The isentropic flow table\n
        
    Examples
    --------
    >>> import gas_dynamics as gd
    >>> gd.stagnation_ratios(range=[0,2], step=.2, gas='nitrogen')  
    M: 0.000   |   P/Pt: 1.000    |    T/Tt: 1.000    |    A/A*: inf  
    M: 0.200   |   P/Pt: 0.972    |    T/Tt: 0.992    |    A/A*: 2.964
    M: 0.400   |   P/Pt: 0.896    |    T/Tt: 0.969    |    A/A*: 1.590
    M: 0.600   |   P/Pt: 0.784    |    T/Tt: 0.933    |    A/A*: 1.188
    M: 0.800   |   P/Pt: 0.656    |    T/Tt: 0.887    |    A/A*: 1.038
    M: 1.000   |   P/Pt: 0.528    |    T/Tt: 0.833    |    A/A*: 1.000
    M: 1.200   |   P/Pt: 0.412    |    T/Tt: 0.776    |    A/A*: 1.030
    M: 1.400   |   P/Pt: 0.314    |    T/Tt: 0.718    |    A/A*: 1.115
    M: 1.600   |   P/Pt: 0.235    |    T/Tt: 0.661    |    A/A*: 1.250
    M: 1.800   |   P/Pt: 0.174    |    T/Tt: 0.607    |    A/A*: 1.439
    M: 2.000   |   P/Pt: 0.128    |    T/Tt: 0.556    |    A/A*: 1.688
    >>>
    """

    gamma = gas.gamma

    Mach_min = range[0]
    Mach_max = range[1]

    mach_nums = [i for i in np.arange(Mach_min,Mach_max+step,step)]
    t_list = [stagnation_temperature_ratio(M=i, gas=gas)for i in mach_nums]
    p_list = [stagnation_pressure_ratio(M=i, gas=gas) for i in mach_nums]
    a_list = [mach_area_ratio_choked(M=i, gas=gas) for i in mach_nums]
    rho_list = [stagnation_density_ratio(M=i, gas=gas) for i in mach_nums]

    labl = '\u03B3 = ' + str(gamma)
    print("Isentropic Flow Parameters for " + gas.name + ", "+ labl)
    for index, num in enumerate(mach_nums):
        print('M: %0.3f' % num, '  |   P/Pt: %0.3f' % p_list[index], '   |    T/Tt: %0.3f' % t_list[index],  '   |    A/A*: %0.3f' % a_list[index],  '   |   rho/rho_t: %0.3f ' % rho_list[index])
    print("\n \n \n")



#==================================================
#mach_from_presure_ratio
#added fluid object
#==================================================    
def mach_from_pressure_ratio(p1: float, p2: float, M1: float, ds=0, gas=air) -> float:
    """Return the Mach number given a Mach number and the local pressures
    
    Notes
    -----
    Given the local pressure in two regions and the Mach number in one,
    return the Mach number in the second region. Default arguments
    are for air and isentropic flow.
    
    Parameters
    ----------
    p1 : `float`
        Pressure in region 1 \n
    p2 : `float`
        Pressure in region 2 \n
    M1 : `float`
        Mach number in region 1 \n
    ds : `float`
        Change in entropy, if any \n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    
    Returns
    -------
    float
        The Mach number\n

    Examples
    --------
    >>> import gas_dynamics as gd
    >>> M2 = gd.mach_from_pressure_ratio(p1=10, p2=2, M1=1) 
    >>> M2
    2.1220079294384067
    >>>
    """

    gamma, R = gas.gamma, gas.R
    M2 = (((p1/p2 * np.exp(ds/R))**((gamma-1)/gamma) * (1 + (gamma-1)/2 * M1**2) - 1) * 2/(gamma-1))**0.5
    return M2



#==================================================
#mach_from_temperature_ratio
#added fluid class
#==================================================
def mach_from_temperature_ratio(T1: float, T2: float, M1: float, gas=air) -> float:
    """Return the Mach number given a Mach number and two local temperatures
    
    Notes
    -----
    Given the local temperatures in two regions and the mach number in one,
    return the Mach number in the second region. Default fluid is air.
    
    Parameters
    ----------
    T1 : `float`
        Temperature in region 1 \n
    T2 : `float`
        Temperature in region 2 \n
    M1 : `float`
        Mach number in region 1 \n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    
    Returns
    -------
    float
        The mach number\n

    Examples
    --------
    >>> import gas_dynamics as gd
    >>> M2 = gd.mach_from_temperature_ratio(T1=300, T2=150, M1=1)
    >>> M2
    2.6457513110645907
    >>>
    """

    gamma = gas.gamma
    M2 = (( T1/T2 * (1 + (gamma-1)/2 * M1**2) - 1) * 2/(gamma-1))**0.5
    return M2



#==================================================
#pressure_from_mach_ratio
#added fluid class
#==================================================
def pressure_from_mach_ratio(M1: float, M2: float, p1: float, ds=0, gas=air) -> float:
    """Return the pressure given a pressure in one region and the two Mach numbers
    
    Notes
    -----
    Given the Mach numbers in two regions and the pressure in one,
    return the missing pressure from the second region. Default arguments
    are for air and isentropic flow.
    
    Parameters
    ----------
    M1 : `float`
        Mach number in region 1 \n
    M2 : `float`
        Mach number in region 2 \n
    p1 : `float`
        Pressure in region 1 \n
    ds : `float`
        Change in entropy, if any \n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    
    Returns
    -------
    float
        The local pressure\n

    Examples
    --------
    >>> import gas_dynamics as gd
    >>> p2 = gd.pressure_from_mach_ratio(M1=1, M2=2, p1=10)  
    >>> p2
    2.4192491286747444
    >>>
    """

    gamma, R = gas.gamma, gas.R
    p2 = p1 * ((1 + ((gamma-1)/2) *M1**2)/(1 + ((gamma-1)/2) *M2**2))**(gamma/(gamma-1)) * np.exp(-ds/R)
    return p2



#==================================================
#temperature_from_mach_ratio
#added fluid class
#==================================================
def temperature_from_mach_ratio(M1: float, M2: float, T1: float, gas=air) -> float:
    """Return the temperature given a temperature in one region and the two Mach numbers
    
    Notes
    -----
    Given the local Mach number in two regions and the temperature in one,
    return the missing temperature from the second region. Default fluid
    is air.
    
    Parameters
    ----------
    M1 : `float`
        Mach number in region 1 \n
    M2 : `float`
        Mach number in region 2 \n
    T1 : `float`
        Temperature in region 1 \n
    gas : `fluid`
        A user defined fluid object. Default is air \n

    Returns
    -------
    float
        The local temperature\n
    
    Examples
    --------
    >>> import gas_dynamics as gd
    >>> T2 = gd.temperature_from_mach_ratio(M1=1, M2=2, T1=297.15) 
    >>> T2
    198.10000000000002
    >>>
    """

    gamma = gas.gamma
    T2 = T1 * (1 + ((gamma-1)/2) *M1**2)/(1 + ((gamma-1)/2) *M2**2)
    return T2



#==================================================
#entropy_produced
#added fluid class
#add examples
#==================================================
def entropy_produced(pt1: float, pt2: float, gas=air) -> float:
    """Return the change in specific entropy from the stagnation pressure ratio
    
    Notes
    -----
    Given two stagnation pressures and the fluid, determine the entropy
    produced per unit mass
    
    Parameters
    ----------
    pt1 : `float`
        Stagnation pressure in region 1 \n
    pt2 : `float`
        Stagnation pressure in region 2 \n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    
    Returns
    -------
    float
        The specific entropy\n
    
    Examples
    --------
    >>> import gas_dynamics as gd
    >>> gd.entropy_produced(pt1=10, pt2=9, gas='air')
    30.238467993796142 #J / kg K
    >>> 
    """
    
    R = gas.R   
    ds = -R*np.log(pt2/pt1)
    return ds



#==================================================
#mach_area_ratio_choked
#added fluid class 
#==================================================
def mach_area_ratio_choked(M: float, gas=air) -> float:
    """Returns the ratio of A / A* given the Mach number.
    
    Notes
    -----
    Given the Mach number and the ratio of specific heats, return the area
    ratio of the Mach number given to the area where Mach number is equal 
    to 1. Default fluid is air.
    
    Parameters
    ----------
    M : `float`
        Mach Number \n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    
    Returns
    -------
    float
        The ratio of area over choked area\n
    
    Examples
    --------
    >>> import gas_dynamics as gd
    >>> A_Astar =gd.mach_area_ratio_choked(M=3)
    >>> A_Astar
    4.23456790123457
    >>>
    """
    if M == 0:
        return float('inf')

    gamma = gas.gamma
    a_star_ratio = 1/M * ((1 + (gamma-1)/2 * M**2) / ((gamma+1)/2)) ** ((gamma+1)/(2*(gamma-1)))
    return a_star_ratio



#==================================================
#mach_area_ratio
#added fluid class
#==================================================
def mach_area_ratio(M1: float, M2: float, gas=air, ds=0) -> float:
    """Return the area ratio given the two Mach numbers

    Notes
    -----
    Given two mach numbers, return the area ratio required to accelerate
    or deaccelerate the flow accordingly. Default fluid is air.
    
    Parameters
    ----------
    M1 : `float` 
        Mach number in region 1 \n
    M2 : `float`
        Mach number in region 2 \n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    ds : `float`
        Entropy produced, if any \n
    
    Returns
    -------
    float
        The area ratio\n

    Examples
    --------
    >>> import gas_dynamics as gd
    >>> A2_A1 = gd.mach_area_ratio(M1=1.5, M2=2.5)
    >>> A2_A1
    2.241789331255894   #area ratio
    >>>
    """

    gamma, R = gas.gamma, gas.R
    A2_A1 = M1/M2 * ((1 + (gamma-1)/2 * M2**2 )/(1 + (gamma-1)/2 * M1**2 ))**((gamma+1)/(2*(gamma-1))) * np.exp(ds/R)
    return A2_A1



#==================================================
#mach_from_area_ratio

#==================================================    
def  mach_from_area_ratio(a_ratio: float, gas=air) ->list:
    """Return the possible mach numbers given a choked area ratio A / A*
    
    Notes
    -----
    Given a ratio of area over an area where Mach = 1, return the subsonic and supersonic
    Mach numbers for the change area. 

    Parameters
    ----------
    a_ratio : `float`
        The ratio of area over choked area \n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    
    Returns
    -------
    list
        The subsonic and supersonic mach numbers for the area ratio\n

    Examples
    --------
    >>> import gas_dynamics as gd
    >>> gd.mach_from_area_ratio(2)
    [0.30590383418910816, 2.197198121652187]
    >>>
    """

    def zero(M, gas):
        return mach_area_ratio_choked(M=M, gas=gas) - a_ratio

    subsonic = fsolve(zero, args=(gas), x0=0.5)
    supersonic = fsolve(zero, args=(gas), x0=5)
    sols = [subsonic[0], supersonic[0]]
    return sols



#==================================================
# mass_flux_max
# added fluid class
#==================================================
def mass_flux_max(pt: float, Tt: float, gas=air, metric=True) -> float:
    """Returns the maximum flow rate per unit choked area
    
    Notes
    -----
    Given stagnation pressure, stagnation temperature, and the fluid, 
    return the flow rate for a Mach number equal to 1. Default fluid 
    is air.

    Units
    -----
    J / kg-K and Pascals return kg/m^2 \n    
    kJ / kg-K and kPa returns kg/m^2 \n  
    ft-lbf / lbm-R and psi returns lbm/in^2 \n  
        
    Parameters
    ----------
    pt : `float`
        The stagnation pressure.\n
    Tt : `float`
        The stagnation temperature.\n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    metric : `bool`
        Use metric or US standard.\n
    
    Returns
    -------
    float
        The mass flux\n
    
    Examples
    --------
    >>> mdot = 5 #kg/s
    >>> mdot_per_area = gd.choked_mdot(1000000, 300) #units are in Pascals
    >>> mdot_per_area
    2333.558560606226
    >>> throat_area = mdot / mdot_per_area
    >>> throat_area             #units are in meters squared
    0.0021426503214477164
    >>>
    #alternatively, we can use the english system; psi, rankine and
    get lbm/s/in^2
    >>> from gas_dynamics.extra import air_us  
    >>> air_us.units
    'Btu / lbm-R'
    >>> flux = gd.mass_flux_max( pt=500, Tt=500, gas=air_us)  
    >>> flux
    2.097208828890205
    >>>
    """

    gamma, R = gas.gamma, gas.R
    if metric == False:
        gc = 32.174 #lbm -ft / lbf -s^2
    else:
        gc = 1
    
    mdot_a_star = (((gc*gamma/(R))*(2/(gamma+1))**((gamma+1)/(gamma-1)))**.5 * pt/(Tt**.5))
    return mdot_a_star



#==================================================
#mass_flux
#added fluid class
#==================================================
def mass_flux(M: float, pt: float, Tt: float, gas = air, metric = True) -> float:
    """Determine mass flow rate for a mach number up to 1

    Notes
    -----
    Given stagnation pressure, stagnation temperature, and the fluid, 
    return the flow rate per unit area for the given Mach number. Default
    fluid is air. 
    
    Units
    -----
    J / kg-K and Pascals return kg/s/m^2 \n    
    kJ / kg-K and kPa returns kg/s/m^2 \n  
    ft-lbf / lbm-R and psi returns lbm/s/in^2 \n  
    Btu / lbm-R and psi returns lbm/s/in^2 \n  
    
    Parameters
    ----------
    M : `float`
        The mach number. Should not exceed 1 \n
    pt : `float`
        The stagnation pressure \n
    Tt : `float`
        The stagnation temperature \n
    gas : `fluid`
        A user defined fluid object. Default is air \n
    
    Returns
    -------
    float
        The mass flux\n

    Examples
    --------
    >>> #metric, input units are Pa, K, output is kg/s/m^2
    >>> flux = gd.mass_flux(M=.8, pt=1e6, Tt=500) 
    >>> flux
    1741.3113452036841
    >>> #us standard, input units are psi and Rankine, output units are lbm/s/in^2
    >>> from gas_dynamics.extra import air_us  
    >>> air_us.units
    'Btu / lbm-R'
    >>> flux = gd.mass_flux(M=.8, pt=500, Tt=500, gas=air_us) 
    >>> flux
    2.01998480961849
    >>>
    """

    gamma, R = gas.gamma, gas.R
    if metric == False:
        gc = 32.174 #lbm-ft / lbf s^2
    else:
        gc = 1
    
    term1 = M * (1 + M**2 * (gamma-1)/2 )**(-(gamma+1)/(2*(gamma-1)))
    mass_flux = term1 * (gamma*gc/R)**.5 * pt/(Tt**.5)
    return mass_flux



#==================================================
#plot_stagnation_ratios
#added fluid object, plot dark and light mode added
#added nice color scheme for lines.
#==================================================
def plot_stagnation_ratios(range=[.1,5], step=.01, gasses=[air, methane, argon], dark=True):
    """Plot the isentropic stagnation relationships for different gasses
    
    Notes
    -----
    Plots Mach number vs T/T, P/Pt, A/A*, rho/rho_t, for a list of
    specific heat ratios.
    
    Parameters
    ----------
    range : `list`
        The starting and ending Mach # in a list, ex: [.01,5] \n
    step : `float`
        The increment between min and max \n
    gasses : `list`
        A list of the user defined gas objects to be plotted \n
        ex: gasses = [air, methane, argon] \n
    dark : `bool`
        Use a dark mode plot. Default true.
    
    Examples
    --------
    >>> import gas_dynamics as gd
    >>> gd.plot_stagnation_ratios()
    """

    if dark == True:
        plt.style.use('dark_background')
        gridcolor = 'w'
    else:
        plt.style.use('default')
        gridcolor = 'k'


    colors = ['tab:blue', 'tab:orange', 'tab:green', 'tab:red', 'tab:purple', 'tab:brown', 'tab:pink', 'tab:gray', 'tab:olive', 'tab:cyan']
    fig, axs = plt.subplots(2,2)
    title = "Isentropic Stagnation Relations"
    fig.suptitle(title)

    Mach_min = range[0]
    Mach_max = range[1]

    for n, f in enumerate(gasses):
        mach_nums = [i for i in np.arange(Mach_min,Mach_max+step,step)]
        t_list = [stagnation_temperature_ratio(M=i, gas=f) for i in mach_nums]
        p_list = [stagnation_pressure_ratio(M=i, gas=f) for i in mach_nums]
        a_list = [mach_area_ratio_choked(M=i, gas=f) for i in mach_nums]
        rho_list = [stagnation_density_ratio(M=i, gas=f) for i in mach_nums]
        labl = '\u03B3 ' + f.name 

    #T/Tt
        axs[0,0].plot(mach_nums,t_list,label=labl, color=colors[n])
        axs[0,0].set_xlabel('Mach Number')
        axs[0,0].set_ylabel('T / Tt')
        axs[0,0].grid(b=True, which='major', color = gridcolor, linestyle = '--', alpha = .1)
        axs[0,0].grid(b=True, which='minor', color = gridcolor, linestyle = '--', alpha = .1)
        axs[0,0].minorticks_on()
        axs[0,0].legend()

    #P/Pt
        axs[0,1].plot(mach_nums,p_list,label=labl, color=colors[n])
        axs[0,1].set_xlabel('Mach Number')
        axs[0,1].set_ylabel('P / Pt')
        axs[0,1].grid(b=True, which='major', color = gridcolor, linestyle = '--', alpha = .1)
        axs[0,1].grid(b=True, which='minor', color = gridcolor, linestyle = '--', alpha = .1)
        axs[0,1].minorticks_on()
        axs[0,1].legend()
    #A/A*
        axs[1,0].plot(mach_nums,a_list,label=labl, color=colors[n])
        axs[1,0].set_xlabel('Mach Number')
        axs[1,0].set_ylabel('A / A*')
        axs[1,0].grid(b=True, which='major', color = gridcolor, linestyle = '--', alpha = .1)
        axs[1,0].grid(b=True, which='minor', color = gridcolor, linestyle = '--', alpha = .1)
        axs[1,0].minorticks_on()
        axs[1,0].legend()

    #rho/rho_t
        axs[1,1].plot(mach_nums,rho_list,label=labl, color=colors[n])
        axs[1,1].set_xlabel('Mach Number')
        axs[1,1].set_ylabel('rho / rho_t')
        axs[1,1].grid(b=True, which='major', color = gridcolor, linestyle = '--', alpha = .1)
        axs[1,1].grid(b=True, which='minor', color = gridcolor, linestyle = '--', alpha = .1)
        axs[1,1].minorticks_on()
        axs[1,1].legend()
    plt.show()


import setuptools
from gas_dynamics.__about__ import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(name="gas_dynamics",
      version=__version__,
      author="Fernando de la Fuente",
      author_email="FernandoAdelaFuente@gmail.com",
      description="Package containing functions for working with compressible flow",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="http://github.com/fernancode/gas_dynamics",
      packages=setuptools.find_packages(),
      classifiers=[
          "Development Status :: 3 - Alpha",
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent"
      ],
      license="MIT",
      python_requires=">=3.6",
      test_suite='nose.collector',
      tests_require=['nose'],
      setup_requires=["numpy==1.16.0"],
      install_requires=["numpy==1.16.0", "scipy", "matplotlib==3.3.2"]
)